
/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "minimumenergyproblem.h"
#include <IPbasic/Errors/mycolordef.h>
#include <OCPbasic/Discretization/trapez.h>
#include <OCPbasic/OCP_Params.h>
#include <OCPbasic/Optimizer/InteriorPointMethod.h>
#include <OCPbasic/solution.h>
#include <iostream>
#include <OCPbasic/linalg/collectormatrix.h>
#include <OCPbasic/linalg/sparsematrix.h>

using IPbasic::TermColor;

int main()
{
    int    NGrid = 100001;
    double t0    = 0.0;
    double tF    = 1.0;
    std::cout << "Construct Minimum Energy Problem..." << std::endl;
    MinimumEnergyProblem ME;
    std::cout << "Minimum Energy Problem constructed!" << std::endl;
    // Startlösung:
    double u0[1];
    u0[0] = -2.0;
    double y0[3];
    for (int k = 0; k < 3; ++k)
        y0[k] = 0.0;
    y0[1]    = 1.0;
    double h = (tF - t0) / ((double)NGrid);

    OCPbasic::Solution<double> Sol(3, 1, 0, NGrid + 1, NGrid + 1);
    Sol.integrateStartSolutionEuler(&ME, h, y0, u0);

    OCPbasic::InteriorPointMethod<double> Opt(&ME, NGrid, t0, tF, OCPbasic::LINSOL_LAPACK_BANDED_KKT);
    OCPbasic::OCP_Params<double> *       Param = Opt.getOCPParameter();
    Param->mu_0                                = 1e-3;
    Param->MyRegularization                    = OCPbasic::REGULARIZE_NONE;
    Param->PrintLev                            = IPbasic::PRINT_MARKED;
    Param->max_iter                            = 1000;

    if (Opt.optimize(Sol))
    {
        TermColor::set_COLOR_GREEN();
        std::cout << "Success!" << std::endl;
        TermColor::set_COLOR_DEFAULT();
    }
    else
    {
        TermColor::set_COLOR_RED();
        std::cout << "Not successfull!" << std::endl;
        TermColor::set_COLOR_DEFAULT();
    }

    Opt.getIPtimings()->printTimeings(std::cout);

    Sol.PrintFile_X("States.txt");
    Sol.PrintFile_U("Controlls.txt");

    return 0;
}
