/* (C) 2018, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file minimumenergyproblem.cpp
 * Implementierung der Klasse MinimumEnergyProblem.
 * Genauere Informationen sind in der Dokumentation der Klasse MinimumEnergyProblem enthalten.
 *
 * @author A. Huber
 * @since 25.09.2018
 */

#include "minimumenergyproblem.h"

MinimumEnergyProblem::MinimumEnergyProblem()
    //                Ny, Nu, Np, N_NLCSTR, N_BOUND
    : Problem<double>(3, 1, 0, 1, 5)
{
    this->setRangeBoxBounds();
}

MinimumEnergyProblem::~MinimumEnergyProblem()
{
}

void MinimumEnergyProblem::setRangeBoxBounds()
{
    // Auf Dimensionen achten!

    // Nichtlineare Nebenbedingung:
    this->g_max[0] = 0.0;
    this->g_min[0] = -OCPbasic::Problem<double>::Inf;
    // this->g_max[0] = 0.0;//Problem<double>::Inf;
    // this->g_min[0] = 0.0;//-Problem<double>::Inf;

    // Randbedingungen:
    // Obere Schranke Randbedingungen:
    this->psi_max[0] = 0.0;
    this->psi_max[1] = 0.0;
    this->psi_max[2] = 0.0;
    this->psi_max[3] = 0.0;
    this->psi_max[4] = 0.0;
    // Untere Schranke Randbedingungen:
    this->psi_min[0] = 0.0;
    this->psi_min[1] = 0.0;
    this->psi_min[2] = 0.0;
    this->psi_min[3] = 0.0;
    this->psi_min[4] = 0.0;

    // Zustände: (Falls nicht überschrieben => Default: -INF < ... < INF)
    // this->y_max[0] = (1.0/9.0);
    // this->y_min[2] = 0.0;
    // Steuerungen: (Falls nicht überschrieben => Default: -INF < ... < INF)

    // Parameter: (Falls nicht überschrieben => Default: -INF < ... < INF)
}

void MinimumEnergyProblem::ODE(double const, double const y[], double const u[], double const[], double f[]) const
{
    f[0] = y[1];
    f[1] = u[0];
    f[2] = 0.5 * u[0] * u[0];
}

void MinimumEnergyProblem::NLCSTR(double const, double const y[], double const[], double const[], double gg[]) const
{
    gg[0] = y[0] - (1.0 / 9.0);
}

void MinimumEnergyProblem::BOUND(const double, const double, const double yF[], const double y0[], const double[], double * psi) const
{
    psi[0] = y0[0];
    psi[1] = yF[0];
    psi[2] = y0[1] - 1.0;
    psi[3] = yF[1] + 1.0;
    psi[4] = y0[2];
}

void MinimumEnergyProblem::VARPHI(const double, const double, const double yF[], const double[], const double[], double phi[]) const
{
    phi[0] = yF[2];
}

//-------------------------------------------------------------------------------------------------------------------------------------//
//                                                                                                                                     //
//                                           Überladen der ersten Ableitungen                                                          //
//                                                                                                                                     //
//-------------------------------------------------------------------------------------------------------------------------------------//

//---------------------------------------------------------------------------------------------------------------------//
//                                                                                                                     //
//                                             Rechte Seite der DGL                                                    //
//                                                                                                                     //
//---------------------------------------------------------------------------------------------------------------------//

void MinimumEnergyProblem::D_ODE_X(double, double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
    // S_OUT.setDimensionen(this->Ny, this->Ny);
    // Hier wird die explizite Ableitung angegeben.
    S_OUT(0, 1) = 1.0;
}

void MinimumEnergyProblem::D_ODE_U(double, double const[], double const u[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
    // S_OUT.setDimensionen(this->Ny, this->Nu);
    // Hier wird die explizite Ableitung angegeben.
    S_OUT(1, 0) = 1.0;
    S_OUT(2, 0) = u[0];
}

void MinimumEnergyProblem::D_ODE_P(double, double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

//---------------------------------------------------------------------------------------------------------------------//
//                                                                                                                     //
//                                        Nichtlineare Nebenbedingungen                                                //
//                                                                                                                     //
//---------------------------------------------------------------------------------------------------------------------//

void MinimumEnergyProblem::D_NLCSTR_X(double, double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
    // S_OUT.setDimensionen(this->N_NLCSTR, this->Ny);
    // Hier wird die explizite Ableitung angegeben.
    S_OUT(0, 0) = 1.0;
}

void MinimumEnergyProblem::D_NLCSTR_U(double, double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void MinimumEnergyProblem::D_NLCSTR_P(double, double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

//---------------------------------------------------------------------------------------------------------------------//
//                                                                                                                     //
//                                               Randbedingungen                                                       //
//                                                                                                                     //
//---------------------------------------------------------------------------------------------------------------------//

void MinimumEnergyProblem::D_BOUND_XF(double, double, double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
    // S_OUT.setDimensionen(this->N_BOUND, this->Ny);
    // Hier wird die explizite Ableitung angegeben.
    S_OUT(1, 0) = 1.0;
    S_OUT(3, 1) = 1.0;
}

void MinimumEnergyProblem::D_BOUND_X0(double, double, double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
    // S_OUT.setDimensionen(this->N_BOUND, this->Ny);
    // Hier wird die explizite Ableitung angegeben.
    S_OUT(0, 0) = 1.0;
    S_OUT(2, 1) = 1.0;
    S_OUT(4, 2) = 1.0;
}

void MinimumEnergyProblem::D_BOUND_P(double, double, double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

//---------------------------------------------------------------------------------------------------------------------//
//                                                                                                                     //
//                                                 Zielfunktion                                                        //
//                                                                                                                     //
//---------------------------------------------------------------------------------------------------------------------//

void MinimumEnergyProblem::D_VARPHI_XF(double, double, double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
    // S_OUT.setDimensionen(1, this->Ny);
    S_OUT(0, 2) = 1.0;
}

void MinimumEnergyProblem::D_VARPHI_X0(double, double, double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void MinimumEnergyProblem::D_VARPHI_P(double, double, double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

//-------------------------------------------------------------------------------------------------------------------------------------//
//                                                                                                                                     //
//                                                    Zweite Ableitungen                                                               //
//                                                                                                                                     //
//-------------------------------------------------------------------------------------------------------------------------------------//

//---------------------------------------------------------------------------------------------------------------------//
//                                                                                                                     //
//                                             Rechte Seite der DGL                                                    //
//                                                                                                                     //
//---------------------------------------------------------------------------------------------------------------------//

void MinimumEnergyProblem::L_DD_ODE_XX(double, double const[], double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void MinimumEnergyProblem::L_DD_ODE_UU(double, double const[], double const[], double const[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
    // Matrix<double> d;
    // Hier wird die explizite Ableitung angegeben.
    double val = lambda[2] * 1.0;
    // S_OUT.setDimensionen(1, 1);
    S_OUT(0, 0) = val;
}

void MinimumEnergyProblem::L_DD_ODE_PP(double, double const[], double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void MinimumEnergyProblem::L_DD_ODE_XU(double, double const[], double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void MinimumEnergyProblem::L_DD_ODE_XP(double, double const[], double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void MinimumEnergyProblem::L_DD_ODE_UP(double, double const[], double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

//---------------------------------------------------------------------------------------------------------------------//
//                                                                                                                     //
//                                        Nichtlineare Nebenbedingungen                                                //
//                                                                                                                     //
//---------------------------------------------------------------------------------------------------------------------//

void MinimumEnergyProblem::L_DD_NLCSTR_XX(double, double const[], double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void MinimumEnergyProblem::L_DD_NLCSTR_UU(double, double const[], double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void MinimumEnergyProblem::L_DD_NLCSTR_PP(double, double const[], double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void MinimumEnergyProblem::L_DD_NLCSTR_XU(double, double const[], double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void MinimumEnergyProblem::L_DD_NLCSTR_XP(double, double const[], double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void MinimumEnergyProblem::L_DD_NLCSTR_UP(double, double const[], double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

//---------------------------------------------------------------------------------------------------------------------//
//                                                                                                                     //
//                                               Randbedingungen                                                       //
//                                                                                                                     //
//---------------------------------------------------------------------------------------------------------------------//

void MinimumEnergyProblem::L_DD_BOUND_XFXF(double, double, double const[], double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void MinimumEnergyProblem::L_DD_BOUND_X0X0(double, double, double const[], double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void MinimumEnergyProblem::L_DD_BOUND_PP(double, double, double const[], double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void MinimumEnergyProblem::L_DD_BOUND_XFX0(double, double, double const[], double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void MinimumEnergyProblem::L_DD_BOUND_XFP(double, double, double const[], double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void MinimumEnergyProblem::L_DD_BOUND_X0P(double, double, double const[], double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

//---------------------------------------------------------------------------------------------------------------------//
//                                                                                                                     //
//                                                 Zielfunktion                                                        //
//                                                                                                                     //
//---------------------------------------------------------------------------------------------------------------------//

void MinimumEnergyProblem::L_DD_VARPHI_XFXF(double, double, double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void MinimumEnergyProblem::L_DD_VARPHI_X0X0(double, double, double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void MinimumEnergyProblem::L_DD_VARPHI_PP(double, double, double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void MinimumEnergyProblem::L_DD_VARPHI_XFX0(double, double, double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void MinimumEnergyProblem::L_DD_VARPHI_XFP(double, double, double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void MinimumEnergyProblem::L_DD_VARPHI_X0P(double, double, double const[], double const[], double const[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}
