/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <IPbasic/Errors/mycolordef.h>
#include <OCPbasic/Discretization/trapez.h>
#include <OCPbasic/Optimizer/InteriorPointMethod.h>
#include <OCPbasic/solution.h>
#include <iostream>
#include <OCPbasic/linalg/collectormatrix.h>
#include <OCPbasic/linalg/sparsematrix.h>
#include "kinmodelocp.h"
#include <fstream>

using IPbasic::TermColor;

int main()
{
    double r     = 0.5;
    double l     = 0.3;
    double c_s   = 0.005;
    int    NGrid = 250;
    double t0    = 0.0;
    double tF    = 1.0;
    std::cout << "Construct simple kinematic OCP..." << std::endl;
    KinModelOCP Kin(l, r, c_s);
    std::cout << "Simple kinematic OCP constructed!" << std::endl;
    // Startlösung:
    double p0[1];
    p0[0] = 1.0;
    double u0[2];
    u0[0] = 0.0;
    u0[1] = 2.0;
    double y0[4];
    for (int k = 0; k < 4; ++k)
        y0[k] = 0.0;
    y0[0] = -1.0;
    double h = (tF - t0) / ((double)NGrid);

    OCPbasic::Solution<double> Sol(4, 2, 1, NGrid + 1, NGrid + 1);
    Sol.integrateStartSolutionEuler(&Kin, h, y0, u0, p0);

    OCPbasic::InteriorPointMethod<double> Opt(&Kin, NGrid, t0, tF,
                                              OCPbasic::LINSOL_LAPACK_BANDED_KKT);
    IPbasic::IPparameter<double> *       Param = Opt.getIPparameter();
    Param->mu_0                                = 1e-3;
    Param->MyLineSearchCouples = IPbasic::COUPLE_STATE_SLACK;

    if (Opt.optimize(Sol))
    {
        TermColor::set_COLOR_GREEN();
        std::cout << "Success!" << std::endl;
        TermColor::set_COLOR_DEFAULT();
    }
    else
    {
        TermColor::set_COLOR_RED();
        std::cout << "Not successfull!" << std::endl;
        TermColor::set_COLOR_DEFAULT();
    }

    double tf = Sol.P[0];

    std::cout << "Optimal solution: " << tf << " sec. to finish manover." << std::endl;

    Sol.PrintFile_TimeScale_X("States.txt", tf);
    Sol.PrintFile_TimeScale_U("Controlls.txt", tf);

    std::ofstream circ;
    circ.open("Circle.txt");
    double hc = 2*M_PI/NGrid;
    for(int i = 0; i <= NGrid; i++)
    {
        double t = hc*i;
        circ << r*sin(t) << '\t' << r*cos(t) << std::endl;
    }
    circ.close();
    return 0;
}
