/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file IPbasic_Errors.cpp
 * Implementation of the static class Errors.
 *
 * @author A. Huber
 * @since 26.10.2016
\*/

#include "IPbasic_Errors.h"
#include "mycolordef.h"
#include <fstream>
#include <iostream>
#include <stdlib.h>
#include <string.h>

namespace IPbasic
{

    // Initialisierung der statischen Variablen:
    bool                                Errors::PrintErrors     = true;
    bool                                Errors::PrintWarnings   = true;
    bool                                Errors::ColoredErrors   = true;
    bool                                Errors::ColoredWarnings = true;
    bool                                Errors::ThrowErrors     = false;
    string_type                         Errors::Akt_Error("No Errors!");
    int_type                            Errors::Error_counter = 0;
    std::vector<Errors::ErrorDataPoint> Errors::MemErrors;
    string_type                         Errors::Akt_Warning("No Warning!");

    int_type                              Errors::Warning_counter = 0;
    std::vector<Errors::WarningDataPoint> Errors::MemWarning;
    MyFunction                            Errors::externalError;   // use default constructor
    MyFunction                            Errors::externalWarning; // use default constructor

    void Errors::set_Error(string_type const & Error)
    {
        Errors::Akt_Error = Error;
        Errors::Error_counter++;

        if (Errors::PrintErrors)
        {
            if (Errors::ColoredErrors)
                TermColor::set_COLOR_RED();
            std::cout << "Error: " << Error << std::endl;
            if (Errors::ColoredErrors)
                TermColor::set_COLOR_DEFAULT();
        }
        Errors::externalError.runExternalFunction(Error);
        if (Errors::ThrowErrors)
            throw std::runtime_error(Error);
        Errors::PushError(Error);
    }

    void Errors::set_Warning(string_type const & Warning)
    {
        Errors::Akt_Warning = Warning;
        Errors::Warning_counter++;

        if (Errors::PrintWarnings)
        {
            if (Errors::ColoredWarnings)
                TermColor::set_COLOR_YELLOW();
            std::cout << "Warning: " << Warning << std::endl;
            if (Errors::ColoredWarnings)
                TermColor::set_COLOR_DEFAULT();
        }
        Errors::externalWarning.runExternalFunction(Warning);
        Errors::PushWarning(Warning);
    }

    int_type Errors::get_Error_Counter()
    {
        return Errors::Error_counter;
    }

    int_type Errors::get_Warning_Counter()
    {
        return Errors::Warning_counter;
    }

    string_type Errors::get_akt_Error_string()
    {
        return Errors::Akt_Error;
    }

    string_type Errors::get_akt_Warning_string()
    {
        return Errors::Akt_Warning;
    }

    std::vector<Errors::ErrorDataPoint> Errors::get_all_Errors()
    {
        return Errors::MemErrors;
    }

    std::vector<Errors::WarningDataPoint> Errors::get_all_Warnings()
    {
        return Errors::MemWarning;
    }

    void Errors::printErrorsToFile(string_type const & FileName)
    {
        std::ofstream File;
        File.open(FileName.data());
        File << "--------------------------------------------" << std::endl;
        if (Errors::Error_counter == 0)
        {
            File << Errors::Akt_Error << std::endl;
        }
        else
        {
            File << "ID" << '\t' << "Count." << '\t' << "Error\n"
                 << "--------------------------------------------" << std::endl;
            for (size_t i = 0; i < Errors::MemErrors.size(); i++)
                File << Errors::MemErrors[i] << std::endl;
        }
        File << "--------------------------------------------" << std::endl;
        if (Errors::Warning_counter == 0)
        {
            File << Errors::Akt_Warning << std::endl;
        }
        else
        {
            File << "ID" << '\t' << "Count." << '\t' << "Warning\n"
                 << "--------------------------------------------" << std::endl;
            for (size_t i = 0; i < Errors::MemWarning.size(); i++)
                File << Errors::MemWarning[i] << std::endl;
        }
        File.close();
    }

    void Errors::resetErrors()
    {
        Errors::Akt_Error     = "No Errors!";
        Errors::Error_counter = 0;
        Errors::MemErrors.clear();
    }

    void Errors::resetWarnings()
    {
        Errors::Akt_Warning     = "No Warnings!";
        Errors::Warning_counter = 0;
        Errors::MemWarning.clear();
    }

    void Errors::PushError(string_type const & Error)
    {
        bool found = false;
        for (size_t i = 0; i < Errors::MemErrors.size(); i++)
        {
            int_type tmp = strcmp(Error.data(), Errors::MemErrors[i].MyError.data());
            if (tmp == 0)
            {
                found = true;
                Errors::MemErrors[i].counter++;
            }
        }
        if (!found)
        {
            Errors::ErrorDataPoint Err;
            Err.MyError = Error;
            Err.counter++;
            Err.Index = int_type(Errors::MemErrors.size());
            Errors::MemErrors.push_back(Err);
        }
    }

    void Errors::PushWarning(string_type const & Warning)
    {
        bool found = false;
        for (size_t i = 0; i < Errors::MemWarning.size(); i++)
        {
            int_type tmp = strcmp(Warning.data(), Errors::MemWarning[i].MyWarning.data());
            if (tmp == 0)
            {
                found = true;
                Errors::MemWarning[i].counter++;
            }
        }
        if (!found)
        {
            Errors::WarningDataPoint Warn;
            Warn.MyWarning = Warning;
            Warn.counter++;
            Warn.Index = int_type(Errors::MemWarning.size());
            Errors::MemWarning.push_back(Warn);
        }
    }

    void Errors::setExternalErrorFunction(void (*function)(string_type const &))
    {
        Errors::externalError.setExternalFunction(function);
    }

    void Errors::setExternalWarningFunction(void (*function)(string_type const &))
    {
        Errors::externalWarning.setExternalFunction(function);
    }

} // namespace IPbasic
