/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file IPbasic_Timings.cpp
 * Implementation of the class IPtimings.
 *
 * @author A. Huber
 * @since 20.05.2019
\*/

#include "IPbasic_Timings.h"

namespace IPbasic
{

    template <typename real>
    IPtimings<real>::IPtimings()
    {
        this->resetTimings();
    }

    template <typename real>
    IPtimings<real>::~IPtimings()
    {
    }

    template <typename real>
    void IPtimings<real>::resetTimings()
    {
        this->TimeOptimize       = real(0.0);
        this->TimeFactorize      = real(0.0);
        this->TimeSolve          = real(0.0);
        this->TimeDataUpdate     = real(0.0);
        this->TimeHessianUpdate  = real(0.0);
        this->TimeJacobianUpdate = real(0.0);
        this->TimeSetupIPSystem  = real(0.0);
        this->TimeLineSearch     = real(0.0);
    }

    template <typename real>
    void IPtimings<real>::printTimeings(ostream_type & stream)
    {
        using std::endl;
        using std::fixed;
        stream << fixed << "Timings IPbasic\n"
               << "TimeOptimize:       " << this->TimeOptimize << "s\n"
               << "TimeLinearSolver:   " << this->TimeFactorize + this->TimeSolve << "s\n"
               << "TimeFactorize:      " << this->TimeFactorize << "s\n"
               << "TimeSolve:          " << this->TimeSolve << "s\n"
               << "TimeDataUpdate:     " << this->TimeDataUpdate << "s\n"
               << "TimeHessianUpdate:  " << this->TimeHessianUpdate << "s\n"
               << "TimeJacobianUpdate: " << this->TimeJacobianUpdate << "s\n"
               << "TimeSetupIPSystem:  " << this->TimeSetupIPSystem << "s\n"
               << "TimeLineSearch:     " << this->TimeLineSearch << "s\n"
               << endl;
    }

    template class IPtimings<float>;
    template class IPtimings<double>;

} // namespace IPbasic
