/* (C) 2016, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file Discretization.cpp
 * Implementation of the class Discretization.
 *
 * @author A. Huber
 * @since 21.04.2016
 */

#include "Discretization.h"

namespace OCPbasic
{

    template <>
    Discretization<double>::Discretization(int_type N, double t0, double tF, Problem<double> * P)
        : N(N)
        , t0(t0)
        , tF(tF)
        , P(P)
        , HesseMatrix(0)
        , JacobiMatrix(0)
    {
        this->MyDims             = new DiscreteDimension<double>(P, this->N, 1E+20, 1E-9);
        this->MyOCP_Params       = new OCP_Params<double>();
        this->MyOCP_Timings      = new OCP_Timings<double>();
        this->N_Statevector      = 0;
        this->N_Constraintvector = 0;
        this->x                  = 0;
        this->lambda             = 0;
        this->h                  = 0.0;
        this->TimeHesse          = 0.0;
        this->TimeJacobi         = 0.0;
        this->TimeIP             = 0.0;
    }

    template <>
    Discretization<float>::Discretization(int_type N, float t0, float tF, Problem<float> * P)
        : N(N)
        , t0(t0)
        , tF(tF)
        , P(P)
        , HesseMatrix(0)
        , JacobiMatrix(0)
    {
        this->MyDims             = new DiscreteDimension<float>(P, this->N, 1E+10, 1E-6);
        this->MyOCP_Params       = new OCP_Params<float>();
        this->MyOCP_Timings      = new OCP_Timings<float>();
        this->N_Statevector      = 0;
        this->N_Constraintvector = 0;
        this->x                  = 0;
        this->lambda             = 0;
        this->h                  = 0.0;
        this->TimeHesse          = 0.0;
        this->TimeJacobi         = 0.0;
        this->TimeIP             = 0.0;
    }

    template <typename real>
    Discretization<real>::~Discretization()
    {
        delete this->MyDims;
        delete this->MyOCP_Params;
        delete this->MyOCP_Timings;
    }

    template class Discretization<double>;
    template class Discretization<float>;

} // namespace OCPbasic
