/* (C) 2019, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file OCP_Timings.cpp
 * Implementation of the class OCP_Timings.
 *
 * @author A. Huber
 * @since 20.05.2019
 */

#include "OCP_Timings.h"

namespace OCPbasic
{

    template <typename real>
    OCP_Timings<real>::OCP_Timings()
        : IPbasic::IPtimings<real>()
    {
        this->resetTimings();
    }

    template <typename real>
    OCP_Timings<real>::~OCP_Timings()
    {
    }

    template <typename real>
    void OCP_Timings<real>::resetTimings()
    {
        // IPbasic Times:
        this->TimeOptimize       = real(0.0);
        this->TimeFactorize      = real(0.0);
        this->TimeSolve          = real(0.0);
        this->TimeDataUpdate     = real(0.0);
        this->TimeHessianUpdate  = real(0.0);
        this->TimeJacobianUpdate = real(0.0);
        this->TimeSetupIPSystem  = real(0.0);
        this->TimeLineSearch     = real(0.0);

        // OCP Times:
        this->TimeInitializeOCP     = real(0.0);
        this->TimeRegularizeHessian = real(0.0);
        this->TimeMatrixConversion  = real(0.0);
    }

    template <typename real>
    void OCP_Timings<real>::printTimeings()
    {
        std::cout << std::fixed;
        std::cout << "Timings IPbasic" << std::endl;
        std::cout << "TimeOptimize:          " << this->TimeOptimize << "s" << std::endl;
        std::cout << "TimeLinearSolver:      " << this->TimeFactorize + this->TimeSolve << "s" << std::endl;
        std::cout << "TimeFactorize:         " << this->TimeFactorize << "s" << std::endl;
        std::cout << "TimeSolve:             " << this->TimeSolve << "s" << std::endl;
        std::cout << "TimeDataUpdate:        " << this->TimeDataUpdate << "s" << std::endl;
        std::cout << "TimeHessianUpdate:     " << this->TimeHessianUpdate << "s" << std::endl;
        std::cout << "TimeJacobianUpdate:    " << this->TimeJacobianUpdate << "s" << std::endl;
        std::cout << "TimeSetupIPSystem:     " << this->TimeSetupIPSystem << "s" << std::endl;
        std::cout << "TimeLineSearch:        " << this->TimeLineSearch << "s" << std::endl;
        std::cout << "TimeInitializeOCP:     " << this->TimeInitializeOCP << "s" << std::endl;
        std::cout << "TimeRegularizeHessian: " << this->TimeRegularizeHessian << "s" << std::endl;
        std::cout << "TimeMatrixConversion:  " << this->TimeMatrixConversion << "s" << std::endl;
        std::cout << std::endl;
    }

    template class OCP_Timings<float>;
    template class OCP_Timings<double>;

} // namespace OCPbasic
