/* (C) 2019, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file ip_solver.h
 * Header definition of the class IP_Solver.
 *
 * @author A. Huber
 * @since 20.02.2019
 */

#ifndef IP_SOLVER_H
#define IP_SOLVER_H

#include "../../OCP_Params.h"
#include "../../OCP_Timings.h"
#include <IPbasic/IPbasic_Solver.h>

namespace OCPbasic
{

    template <typename real>
    class IP_Solver : public IPbasic::Solver<real>
    {
      public:
        /**
         * \brief IP_Solver:
         *        Contructor of the class IP_Solver.
         *
         * \param[in] Problem               NLP-Problem, which has to be solved.
         * \param[in] LinSolver             Solver of the linear system.
         * \param[in] IPParams              Interior-point-parameters.
         * \param[in] Timings               Timings of the interior-point-method.
         * \param[in] UsedLineSearchMethod  Used LineSearchMethod.
         *
         */
        IP_Solver(IPbasic::NLP<real> *                    Problem,
                  IPbasic::LinearSolver_interface<real> * LinSolver,
                  OCP_Params<real> *                      OCPParams,
                  OCP_Timings<real> *                     Timings,
                  IPbasic::LineSearchMethod               UsedLineSearchMethod = IPbasic::LINESEARCH_ARMIJO_L1_PENALTY);

        /**
         * \brief ~IP_Solver:
         *        Virtual destructor of the class IP_Solver.
         */
        virtual ~IP_Solver();
    };

} // namespace OCPbasic

#endif // IP_SOLVER_H
