/* (C) 2019, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "DenseLU.h"

#if defined(OCPBASIC_USE_ALGLIN)

#else

template <typename real>
DenseLU<real>::DenseLU(int_type DimRHS)
    : mem_real("DenseLUReal")
    , mem_int("DenseLUIntType")
    , DimRHS(DimRHS)
{
    this->LDA = DimRHS;
    this->mem_real.allocate(this->LDA * this->DimRHS);
    this->mem_int.allocate(this->DimRHS);
    this->A    = this->mem_real(size_t(this->LDA * this->DimRHS));
    this->IPIV = this->mem_int(size_t(this->DimRHS));
}

template <typename real>
DenseLU<real>::~DenseLU()
{
    this->mem_real.free();
    this->mem_int.free();
}

template <typename real>
real * DenseLU<real>::getValueArray()
{
    return this->A;
}

template <typename real>
int_type DenseLU<real>::getDimension()
{
    return this->DimRHS;
}

template <typename real>
bool DenseLU<real>::factorize()
{
    int_type info = this->getrf(this->DimRHS, this->DimRHS, this->A, this->LDA, this->IPIV);
    if (info == 0)
        return true;
    else
        return false;
}

template <typename real>
bool DenseLU<real>::solve(real RHS[])
{
    int_type info = this->getrs('N', this->DimRHS, 1, this->A, this->LDA, this->IPIV, RHS, this->DimRHS);
    if (info == 0)
        return true;
    else
        return false;
}

template <typename real>
bool DenseLU<real>::solve(int_type NRHS, real RHS[])
{
    int_type info = this->getrs('N', this->DimRHS, NRHS, this->A, this->LDA, this->IPIV, RHS, this->DimRHS);
    if (info == 0)
        return true;
    else
        return false;
}

template <typename real>
void DenseLU<real>::clear()
{
    std::fill(this->A, this->A + size_t(this->LDA * this->DimRHS), real(0.0));
}

template <typename real>
int_type DenseLU<real>::getrf(int_type M, int_type N, real A[], int_type LDA, int_type IPIV[])
{
    int_type INFO = 0;
    INFO          = lapack_wrapper::getrf(M, N, A, LDA, IPIV);
    return INFO;
}

template <typename real>
int_type DenseLU<real>::getrs(char TRANS, int_type N, int_type NRHS, real A[], int_type LDA, int_type IPIV[], real B[], int_type LDB)
{
    int_type                      INFO = 0;
    lapack_wrapper::Transposition tra;
    if (TRANS == 'N')
    {
        tra = lapack_wrapper::NO_TRANSPOSE;
    }
    else if (TRANS == 'T')
    {
        tra = lapack_wrapper::TRANSPOSE;
    }
    else if (TRANS == 'C')
    {
        tra = lapack_wrapper::CONJUGATE_TRANSPOSE;
    }
    else
    {
        tra = lapack_wrapper::NO_TRANSPOSE;
    }
    INFO = lapack_wrapper::getrs(tra, N, NRHS, A, LDA, IPIV, B, LDB);
    return INFO;
}

template class DenseLU<double>;
template class DenseLU<float>;

#endif
