/* (C) 2019, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file OCPbasic.cpp
 * Implementation of the basic mathods in OCPbasic.
 *
 * @author A. Huber
 * @since 01.08.2019
 */

#include "OCPbasic.h"
#include <IPbasic/Errors/IPbasic_Errors.h>

namespace OCPbasic
{

    template <typename real>
    void CopyRow(IPbasic::MatrixWrapper<real> const & From, IPbasic::int_type FromRow, IPbasic::MatrixWrapper<real> & To, IPbasic::int_type ToRow)
    {
        if (From.numCols() != To.numCols() || ToRow >= To.numRows() || FromRow >= From.numRows())
        {
            IPbasic::Errors::set_Error("CopyRow bad dimensions!");
            return;
        }
        for (IPbasic::int_type j = 0; j < From.numCols(); ++j)
            To(ToRow, j) = From(FromRow, j);
    }

    template <typename real>
    void CopyNegRow(IPbasic::MatrixWrapper<real> const & From, IPbasic::int_type FromRow, IPbasic::MatrixWrapper<real> & To, IPbasic::int_type ToRow)
    {
        if (From.numCols() != To.numCols() || ToRow >= To.numRows() || FromRow >= From.numRows())
        {
            IPbasic::Errors::set_Error("CopyNegRow bad dimensions!");
            return;
        }
        for (IPbasic::int_type j = 0; j < From.numCols(); ++j)
            To(ToRow, j) = -From(FromRow, j);
    }

    template void CopyRow(IPbasic::MatrixWrapper<float> const & From, IPbasic::int_type FromRow, IPbasic::MatrixWrapper<float> & To, IPbasic::int_type ToRow);

    template void CopyRow(IPbasic::MatrixWrapper<double> const & From, IPbasic::int_type FromRow, IPbasic::MatrixWrapper<double> & To, IPbasic::int_type ToRow);

    template void CopyNegRow(IPbasic::MatrixWrapper<float> const & From, IPbasic::int_type FromRow, IPbasic::MatrixWrapper<float> & To, IPbasic::int_type ToRow);

    template void CopyNegRow(IPbasic::MatrixWrapper<double> const & From, IPbasic::int_type FromRow, IPbasic::MatrixWrapper<double> & To, IPbasic::int_type ToRow);

} // namespace OCPbasic
