#ifndef IPBASIC_LAPACK_BLAS_H
#define IPBASIC_LAPACK_BLAS_H

#include <IPbasic/IPbasic.h>

using IPbasic::int_type;

#ifndef LAPACK_WRAPPER_USE_MKL

#ifdef __cplusplus
extern "C"
{
#endif

    /**
     * \brief dsyev_: (LAPACK double)
     *
     *
     * \param[in]     JOBZ      = 'N':  Compute eigenvalues only;
     *                          = 'V':  Compute eigenvalues and eigenvectors.
     * \param[in]     UPLO      = 'U':  Upper triangle of A is stored;
     *                          = 'L':  Lower triangle of A is stored.
     * \param[in]     N         The order of the matrix A.  N >= 0.
     * \param[in,out] A         Value array of A.
     * \param[in]     LDA       Leading dimension of A.
     * \param[out]    W         If INFO = 0, the eigenvalues in ascending order.
     * \param[out]    WORK      Workarray. On exit, if INFO = 0, WORK(1) returns the optimal LWORK.
     * \param[in]     LWORK     The length of the array WORK.  LWORK >= max(1,3*N-1).
     *                          For optimal efficiency, LWORK >= (NB+2)*N,
     *                          where NB is the blocksize for DSYTRD returned by ILAENV.
     *                          If LWORK = -1, then a workspace query is assumed; the routine
     *                          only calculates the optimal size of the WORK array, returns
     *                          this value as the first entry of the WORK array, and no error
     *                          message related to LWORK is issued by XERBLA.
     * \param[out]    INFO      = 0:  successful exit
     *                          < 0:  if INFO = -i, the i-th argument had an illegal value
     *                          > 0:  if INFO = i, the algorithm failed to converge; i
     *                                off-diagonal elements of an intermediate tridiagonal
     *                                form did not converge to zero.
     *
     */
    void dsyev_(char * JOBZ, char * UPLO, int_type * N, double A[], int_type * LDA, double W[], double WORK[], int_type * LWORK, int_type * INFO);

    /**
     * \brief ssyev_: (LAPACK float)
     *
     *
     * \param[in]     JOBZ      = 'N':  Compute eigenvalues only;
     *                          = 'V':  Compute eigenvalues and eigenvectors.
     * \param[in]     UPLO      = 'U':  Upper triangle of A is stored;
     *                          = 'L':  Lower triangle of A is stored.
     * \param[in]     N         The order of the matrix A.  N >= 0.
     * \param[in,out] A         Value array of A.
     * \param[in]     LDA       Leading dimension of A.
     * \param[out]    W         If INFO = 0, the eigenvalues in ascending order.
     * \param[out]    WORK      Workarray. On exit, if INFO = 0, WORK(1) returns the optimal LWORK.
     * \param[in]     LWORK     The length of the array WORK.  LWORK >= max(1,3*N-1).
     *                          For optimal efficiency, LWORK >= (NB+2)*N,
     *                          where NB is the blocksize for DSYTRD returned by ILAENV.
     *                          If LWORK = -1, then a workspace query is assumed; the routine
     *                          only calculates the optimal size of the WORK array, returns
     *                          this value as the first entry of the WORK array, and no error
     *                          message related to LWORK is issued by XERBLA.
     * \param[out]    INFO      = 0:  successful exit
     *                          < 0:  if INFO = -i, the i-th argument had an illegal value
     *                          > 0:  if INFO = i, the algorithm failed to converge; i
     *                                off-diagonal elements of an intermediate tridiagonal
     *                                form did not converge to zero.
     *
     */
    void ssyev_(char * JOBZ, char * UPLO, int_type * N, float A[], int_type * LDA, float W[], float WORK[], int_type * LWORK, int_type * INFO);

#ifdef __cplusplus
}
#endif

#endif // LAPACK_WRAPPER_USE_MKL

#endif // IPBASIC_LAPACK_BLAS_H
