/* (C) 2018, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file solution.cpp
 * Implementation of the class Solution.
 *
 * @author A. Huber
 * @since 07.11.2018
 */

#include "solution.h"
#include "problem.h"
#include <IPbasic/Errors/IPbasic_Errors.h>
#include <fstream>

namespace OCPbasic
{

    template <typename real>
    Solution<real>::Solution(int_type dim_X, int_type dim_U, int_type dim_P, int_type grid_X, int_type grid_U)
        : dim_X(dim_X)
        , dim_U(dim_U)
        , dim_P(dim_P)
        , grid_X(grid_X)
        , grid_U(grid_U)
        , isSet(false)
        , isStKonst(false)
        , isMaxIter(false)
    {
        // Allokiere Speicher:
        this->X   = new real *[this->grid_X];
        this->U   = new real *[this->grid_U];
        this->t_X = new real[this->grid_X];
        this->t_U = new real[this->grid_U];
        for (int_type i = 0; i < this->grid_X; ++i)
            this->X[i] = new real[this->dim_X];
        for (int_type i = 0; i < this->grid_U; ++i)
            this->U[i] = new real[this->dim_U];
        this->P = new real[this->dim_P];
    }

    template <typename real>
    Solution<real>::~Solution()
    {
        // Deallokiere Speicher:
        for (int_type i = 0; i < this->grid_X; ++i)
            delete[] this->X[i];
        for (int_type i = 0; i < this->grid_U; ++i)
            delete[] this->U[i];
        delete[] this->X;
        delete[] this->U;
        delete[] this->t_X;
        delete[] this->t_U;

        delete[] this->P;
    }

    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                     Public Memberfunktionen                                                         //
    //                                                                                                                                     //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    template <typename real>
    void Solution<real>::PrintFile_X(std::string const & file)
    {
        std::string tab;
        if (this->testFileType(file, std::string(".csv")))
            tab = ",   ";
        else
            tab = "    ";
        if (this->isSet)
        {
            std::ofstream myfile;
            myfile.open(file.data());
            for (int_type i = 0; i < this->grid_X; ++i)
            {
                myfile << this->t_X[i];
                for (int_type j = 0; j < this->dim_X; ++j)
                    myfile << tab << this->X[i][j];
                myfile << std::endl;
            }
            myfile.close();
        }
        else
        {
            IPbasic::Errors::set_Error("Ausgabe der Lösung in ein File nicht möglich, da keine Lösung vorhanden!");
        }
    }

    template <typename real>
    void Solution<real>::PrintFile_U(std::string const & file)
    {
        std::string tab;
        if (this->testFileType(file, std::string(".csv")))
            tab = ",   ";
        else
            tab = "    ";
        if (this->isSet)
        {
            if (this->isStKonst)
            {
                real          h = this->t_U[1] - this->t_U[0];
                std::ofstream myfile;
                myfile.open(file.data());
                for (int_type i = 0; i < this->grid_U; ++i)
                {
                    myfile << this->t_U[i];
                    for (int_type j = 0; j < this->dim_U; ++j)
                        myfile << tab << this->U[i][j];
                    myfile << std::endl;
                    myfile << this->t_U[i] + h;
                    for (int_type j = 0; j < this->dim_U; ++j)
                        myfile << tab << this->U[i][j];
                    myfile << std::endl;
                }
                myfile.close();
            }
            else
            {
                std::ofstream myfile;
                myfile.open(file.data());
                for (int_type i = 0; i < this->grid_U; ++i)
                {
                    myfile << this->t_U[i];
                    for (int_type j = 0; j < this->dim_U; ++j)
                        myfile << tab << this->U[i][j];
                    myfile << std::endl;
                }
                myfile.close();
            }
        }
        else
        {
            IPbasic::Errors::set_Error("Ausgabe der Lösung in ein File nicht möglich, da keine Lösung vorhanden!");
        }
    }

    template <typename real>
    void Solution<real>::PrintFile_P(std::string const & file)
    {
        std::string tab;
        if (this->testFileType(file, std::string(".csv")))
            tab = ",   ";
        else
            tab = "    ";
        if (this->isSet)
        {
            std::ofstream myfile;
            myfile.open(file.data());
            for (int_type j = 0; j < this->dim_P; ++j)
                myfile << tab << this->P[j];
            myfile.close();
        }
        else
        {
            IPbasic::Errors::set_Error("Ausgabe der Lösung in ein File nicht möglich, da keine Lösung vorhanden!");
        }
    }

    template <typename real>
    void Solution<real>::PrintFile_XUP(std::string const & file)
    {
        if (this->grid_X != this->grid_U)
        {
            IPbasic::Errors::set_Error("PrintFile_XUP: Cannot print into file because grid_X != this->grid_U!");
            return;
        }
        std::string tab;
        if (this->testFileType(file, std::string(".csv")))
            tab = ",   ";
        else
            tab = "    ";
        if (this->isSet)
        {
            std::ofstream myfile;
            myfile.open(file.data());
            for (int_type i = 0; i < this->grid_X; ++i)
            {
                myfile << this->t_X[i];
                for (int_type j = 0; j < this->dim_X; ++j)
                    myfile << tab << this->X[i][j];
                for (int_type j = 0; j < this->dim_U; ++j)
                    myfile << tab << this->U[i][j];
                for (int_type j = 0; j < this->dim_P; ++j)
                    myfile << tab << this->P[j];
                myfile << std::endl;
            }
            myfile.close();
        }
        else
        {
            IPbasic::Errors::set_Error("Ausgabe der Lösung in ein File nicht möglich, da keine Lösung vorhanden!");
        }
    }

    template <typename real>
    void Solution<real>::PrintFile_TimeScale_X(std::string const & file, real scale)
    {
        std::string tab;
        if (this->testFileType(file, std::string(".csv")))
            tab = ",   ";
        else
            tab = "    ";
        if (this->isSet)
        {
            std::ofstream myfile;
            myfile.open(file.data());
            for (int_type i = 0; i < this->grid_X; ++i)
            {
                myfile << this->t_X[i] * scale;
                for (int_type j = 0; j < this->dim_X; ++j)
                    myfile << tab << this->X[i][j];
                myfile << std::endl;
            }
            myfile.close();
        }
        else
        {
            IPbasic::Errors::set_Error("Ausgabe der Lösung in ein File nicht möglich, da keine Lösung vorhanden!");
        }
    }

    template <typename real>
    void Solution<real>::PrintFile_TimeScale_U(std::string const & file, real scale)
    {
        std::string tab;
        if (this->testFileType(file, std::string(".csv")))
            tab = ",   ";
        else
            tab = "    ";
        if (this->isSet)
        {
            if (this->isStKonst)
            {
                real          h = this->t_U[1] - this->t_U[0];
                std::ofstream myfile;
                myfile.open(file.data());
                for (int_type i = 0; i < this->grid_U; ++i)
                {
                    myfile << this->t_U[i] * scale;
                    for (int_type j = 0; j < this->dim_U; ++j)
                        myfile << tab << this->U[i][j];
                    myfile << std::endl;
                    myfile << (this->t_U[i] + h) * scale;
                    for (int_type j = 0; j < this->dim_U; ++j)
                        myfile << tab << this->U[i][j];
                    myfile << std::endl;
                }
                myfile.close();
            }
            else
            {
                std::ofstream myfile;
                myfile.open(file.data());
                for (int_type i = 0; i < this->grid_U; ++i)
                {
                    myfile << this->t_U[i] * scale;
                    for (int_type j = 0; j < this->dim_U; ++j)
                        myfile << tab << this->U[i][j];
                    myfile << std::endl;
                }
                myfile.close();
            }
        }
        else
        {
            IPbasic::Errors::set_Error("Ausgabe der Lösung in ein File nicht möglich, da keine Lösung vorhanden!");
        }
    }

    template <typename real>
    void Solution<real>::PrintFile_TimeScale_XUP(std::string const & file, real scale)
    {
        if (this->grid_X != this->grid_U)
        {
            IPbasic::Errors::set_Error("PrintFile_XUP: Cannot print into file because grid_X != this->grid_U!");
            return;
        }
        std::string tab;
        if (this->testFileType(file, std::string(".csv")))
            tab = ",   ";
        else
            tab = "    ";
        if (this->isSet)
        {
            std::ofstream myfile;
            myfile.open(file.data());
            for (int_type i = 0; i < this->grid_X; ++i)
            {
                myfile << this->t_X[i] * scale;
                for (int_type j = 0; j < this->dim_X; ++j)
                    myfile << tab << this->X[i][j];
                for (int_type j = 0; j < this->dim_U; ++j)
                    myfile << tab << this->U[i][j];
                for (int_type j = 0; j < this->dim_P; ++j)
                    myfile << tab << this->P[j];
                myfile << std::endl;
            }
            myfile.close();
        }
        else
        {
            IPbasic::Errors::set_Error("Ausgabe der Lösung in ein File nicht möglich, da keine Lösung vorhanden!");
        }
    }

    template <typename real>
    void Solution<real>::integrateStartSolutionEuler(Problem<real> const * OCP, real h, real const y0[], real const u[], real const p0[])
    {
        // Teste Dimensionen:
        if (OCP->Ny != this->dim_X || OCP->Nu != this->dim_U || OCP->Np != this->dim_P)
        {
            IPbasic::Errors::set_Error("Dimensionen stimmen nicht überein!");
            return;
        }
        // Belege u:
        for (int_type i = 0; i < this->grid_U; ++i)
            memcpy(this->U[i], u, this->dim_U * sizeof(real));
        // Integriere y:
        // y0:
        memcpy(this->X[0], y0, this->dim_X * sizeof(real));
        real * f = new real[this->dim_X];
        real * y_i;
        real * y_i_;
        real   t_i   = 0.0;
        this->t_X[0] = 0.0;
        this->t_U[0] = 0.0;
        for (int_type i = 1; i < this->grid_X; ++i)
        {
            // Eulerintegration:
            t_i += h;
            this->t_X[i] = t_i;
            this->t_U[i] = t_i;
            y_i          = this->X[i];
            y_i_         = this->X[i - 1];
            OCP->ODE(t_i, y_i_, u, p0, f);
            // Belege y_i:
            for (int_type j = 0; j < this->dim_X; ++j)
                y_i[j] = y_i_[j] + h * f[j];
        }
        delete[] f;
        // Initialisiere Parameter:
        if (p0 == 0 && this->dim_P > 0)
        {
            IPbasic::Errors::set_Warning("Keine Parameter Startschätzung gegeben!");
            for (int_type i = 0; i < this->dim_P; ++i)
                this->P[i] = 0.0;
        }
        else
        {
            memcpy(this->P, p0, this->dim_P * sizeof(real));
        }
        this->isSet = true;
    }

    template <typename real>
    void Solution<real>::integrateStartSolutionEuler(Problem<real> const * OCP, real const t[], real const y0[], real const u[], real const p0[])
    {
        // Teste Dimensionen:
        if (OCP->Ny != this->dim_X || OCP->Nu != this->dim_U || OCP->Np != this->dim_P)
        {
            IPbasic::Errors::set_Error("Dimensionen stimmen nicht überein!");
            return;
        }
        // Belege u:
        for (int_type i = 0; i < this->grid_U; ++i)
            memcpy(this->U[i], u, this->dim_U * sizeof(real));
        // Integriere y:
        // y0:
        memcpy(this->X[0], y0, this->dim_X * sizeof(real));
        real * f = new real[this->dim_X];
        real * y_i;
        real * y_i_;
        real   t_i;
        real   h;
        this->t_X[0] = 0.0;
        this->t_U[0] = 0.0;
        for (int_type i = 1; i < this->grid_X; ++i)
        {
            // Eulerintegration:
            h            = t[i] - t[i - 1];
            t_i          = t[i];
            this->t_X[i] = t_i;
            this->t_U[i] = t_i;
            y_i          = this->X[i];
            y_i_         = this->X[i - 1];
            OCP->ODE(t_i, y_i_, u, p0, f);
            // Belege y_i:
            for (int_type j = 0; j < this->dim_X; ++j)
                y_i[j] = y_i_[j] + h * f[j];
        }
        delete[] f;
        // Initialisiere Parameter:
        if (p0 == 0 && this->dim_P > 0)
        {
            IPbasic::Errors::set_Warning("Keine Parameter Startschätzung gegeben!");
            for (int_type i = 0; i < this->dim_P; ++i)
                this->P[i] = 0.0;
        }
        else
        {
            memcpy(this->P, p0, this->dim_P * sizeof(real));
        }
        this->isSet = true;
    }

    template <typename real>
    bool Solution<real>::testFileType(std::string FilePath, std::string Type)
    {
        int_type LenFilePath = FilePath.length();
        int_type LenType     = Type.length();
        if (LenFilePath < LenType)
        {
            return false;
        }
        else
        {
            bool ret = true;
            for (int_type i = 0; i < LenType; i++)
            {
                if (FilePath.data()[LenFilePath - 1 - i] != Type.data()[LenType - 1 - i])
                {
                    ret = false;
                }
            }
            return ret;
        }
    }

    template class Solution<double>;
    template class Solution<float>;

} // namespace OCPbasic
